//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//             Box Sampling
// 
// Filename:   box.c, version 1.03, 3 April 2004
//
// Purpose:    Bayesian computation.
//
//                    Pr(x)  Pr(D|x)  =   Pr(D)  Pr(x|D)
//                   Prior.Likelihood = Evidence.Posterior
//
//             Obtain Evidence and quantified properties of x,
//             with posterior samples as optional by-product.
//
//             John Skilling, Kenmare, Ireland, April 2004
//             email: skilling@eircom.net
//=============================================================================
/*
    Copyright (c) 2004, Maximum Entropy Data Consultants Ltd,
                        114c Milton Road, Cambridge CB4 1XE, England

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation Inc.,
    59 Temple Place, Suite 330, Boston, MA  02111-1307  USA; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.
*/
//=============================================================================
//
//                                MAIN
//                                  |
//                                  |
//                              BoxSampler
//            _____________________/|\___________________________
//           |            |         |          |                 |
//        BoxCopy   BoxProperties   |       BoxReset        BoxPosterior
//           |                      |       BoxReset1            |
//           |                      |                            |
//           |          ___________/|\_______________________    |
//           |         |            |        |               |   |
//           |      UserPrior  UserExplore  UserProperties  UserCopy
//           |                      |         
//           |                      |         
//           |           __________/|\__________
//           |          |           |           |
//           |      BoxLhood    BoxTarget     BoxCorr
//           |       |
//           BoxLtrees
//
//-----------------------------------------------------------------------------
#include <stdlib.h>
#include <stdarg.h>
#include <float.h>
#include <math.h>
#include "userstr.h"              // Your header, needed before box.h
#include "box.h"                  // My header, needs definitions in userstr.h
#include "random.h"               // Random library

/***********************/
/* Internal prototypes */
/***********************/
static void     BoxCopy      (GlobalStr*, int, int);
static void     BoxLtrees    (int, double, double, double*, double*, double*);
static int      BoxReset     (GlobalStr*);
static void     BoxReset1    (int, int, double*);
static int      BoxPosterior (GlobalStr*, ItemStr*, int, ItemStr*);
static void     BoxProperties(GlobalStr*);

/*********************************/
/* Internal constants and macros */
/*********************************/
#undef  CALLOC    // allocates vector p[0:n-1] of type t
#undef  FREE      // frees CALLOC or NULL vector p[0:*], sets NULL
#undef  CALL      // call mechanism enabling clean abort if return code < 0
#undef  MIN       // minimum
#undef  MAX       // maximum
#undef  PLUS      // x+y when numbers are stored as logarithms
#undef  GT        // L1>L2 for extended values

#define CALLOC(p,n,t) {p=NULL;\
if((n)>0&&!(p=(t*)calloc((size_t)(n),sizeof(t))))\
{CALLvalue=E_MALLOC;goto Exit;}/*printf("%p %d\n",p,(size_t)(n)*sizeof(t));*/}
#define FREE(p) {if(p){/*printf("%p -1\n",p);*/(void)free((void*)p);} p=NULL;}
#define CALL(x)    {if( (CALLvalue = (x)) < 0 ) goto Exit;}
#define MIN(x,y)   ((x) < (y) ? (x) : (y))
#define MAX(x,y)   ((x) > (y) ? (x) : (y))
#define PLUS(x,y)  ((x)>(y)?(x)+log(1.+exp((y)-(x))):(y)+log(1.+exp((x)-(y))))
#define GT(L1,e1, L2,e2) (L1>L2 || (L1==L2 && e1>e2))

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function:  BoxSampler
// 
// Purpose:   Bayesian computation
//
//                   Pr(x)  Pr(D|x)   =    Pr(D)  Pr(x|D)
//                  Prior Likelihood ==> Evidence Posterior
//
//            Use "BoxSampler" to obtain
//
//               lnEvidence  := ln INTEGRAL Likelihood d(Prior),
//               devlnEvid   := std.dev. uncertainty of lnEvidence
//
//               Information := ln( prior-to-posterior compression factor )
//            
//               Number      := effective # independent samples during run
//
//               Nactive     := apparent # active constraints
//
//               Posterior samples (optional)
//
//               Posterior mean properties (optional)
//
// History:   John Skilling   3 April 2004
//-----------------------------------------------------------------------------
int BoxSampler(        //   O  last UserProperties return code, or -ve error
GlobalStr* Global,     // I O  general information, see structure definition
ItemStr*   Items,      //   O  item collection               [Global->NumItems]
ItemStr*   Samples,    //  (O) posterior samples                [<= MaxSamples]
int        MaxSamples, // I    max # posterior samples          (0 allowed)
int        MaxSteps,   // I    max # steps allowed            (-ve is inactive)
double     tol)        // I    fractional tolerance on Evidence (0 is inactive)
{
static const double logZERO = -DBL_MAX * DBL_EPSILON;   // ln(0) = -infinity
static const double PIE     = 8.53973422267356706540;   // pi * e
    int       N    = Global->NumItems;
    double    c1   = log(N / (N + 1.0));   // ln recurrence factor
    double    c2   = log(N / (N + 2.0));   // ln recurrence factor
    double    L;             // ln Likelihood
    double    h;             // ln current interval
    double    a;             // ln recurrence term
    double    b;             // ln recurrence term
    double    Evid;          // ln accumulated Evidence
    double    Evid2;         // ln accumulated Evidence^2
    double    Info;          // <ln L>
    double    Logw;          // <ln w>
    double    wnew;          // proportional weight of new item
    double    wold;          // proportional weight of accumulated old items
    double    Number;        // Effective # independent items
    int       edge;          // &(item with minimum L on entry)
    int       copy;          // &(duplication item)
    int       i;             // counters
    int       Nplus;         // power-of-2 >= N
    int       CALLvalue = 0;

// Setup
    a = log(2.0 / ((N + 1.0) * N));
    h = log(1.0 / N);
    Evid = Evid2 = b = logZERO;
    Global->Nsamples = Global->Nsteps = 0;
    Info = Logw = 0.0;
    CALL( RanInit(Global->Rand, Global->Iseed) )
    Global->Iseed = CALLvalue;
    RanInit(Global->Rpost, Global->Iseed | 1);
    for( Nplus = 1; Nplus < N; Nplus <<= 1 ) ;
    Global->Nplus = Nplus;
    Global->CorrTree = NULL;
    Global->TopTree  = NULL;
    Global->EdgeTree = NULL;
    CALLOC(Global->CorrTree, 2 * Nplus, double)   // Corr[-Nplus:Nplus-1]
    CALLOC(Global->TopTree,  2 * Nplus, double)   // Lmax[-Nplus:Nplus-1]
    CALLOC(Global->EdgeTree, 2 * Nplus, double)   // Lmin[-Nplus:Nplus-1]
    CALLOC(Global->extTree,  2 * Nplus, double)   //  ext[-Nplus:Nplus-1]
    for( i = 0; i < Nplus * 2; i++ )
    {
        Global->CorrTree[i] = 0.0;            // Corr = 0
        Global->TopTree[i]  = logZERO;        // Lmax = -infinity
        Global->EdgeTree[i] = -logZERO;       // Lmin = +infinity
        Global->extTree[i]  = 1.0;            // Lmin extension = full
    }
    Global->Corr  = Global->CorrTree + Nplus; // Corr [0:N-1] for diagnostic
    Global->Lhood = Global->EdgeTree + Nplus; // Lhood[0:N-1] for diagnostic
    for( i = 0; i < Global->NumProps; i++ )
        Global->Properties[i] = Global->Mean[i] = Global->Dev[i] = 0.0;
    Global->lnLmin = logZERO;
    Global->ext = 0.0;

// Prior
    for( i = 0; i < N; i++ )
    {
        CALL( UserPrior(Global, &Items[i], &L) )
        BoxLhood(Global, 1, i, L);
    }
    edge = BoxReset(Global);

// Convergence steps
    while( Global->Nsteps++ != MaxSteps )
    {
//  locate edge
        L = Global->lnLmin;
//  update system statistics
        h += c1;
        a += c2;
        b = PLUS(b + c1, L + a);
        Evid  = PLUS(Evid,  h + L);
        Evid2 = PLUS(Evid2, b + L);
        Global->wnew = wnew = exp(h + L - Evid);
        Global->wold = wold = MAX(1.0 - wnew, 0.0);
        Info   = wold * Info + wnew * L;
        Logw   = wold * Logw + wnew * (h + L);
        Number = exp(Evid - Logw);
        Global->lnEvidence  = 2.0 * Evid - 0.5 * Evid2;
        Global->devlnEvid   = sqrt(MAX(Evid2 - 2.0 * Evid, 0.0));
        Global->Information = Info - Evid;
        Global->Number  = (int)(0.5 + Number);
        Global->Nactive = (int)(0.5 + Number * Number / (PIE * N * N));
//  optional samples and user statistics
//    for( i = 0; i < N; i++ ) printf("%7.4f", Items[i].x); printf("\n");
        CALL( BoxPosterior(Global, Samples, MaxSamples, &Items[edge]) )
        CALL( UserProperties(Global, &Items[edge], Samples) )
        BoxProperties(Global);
        if( N > 1 )
        {
//  new item within box L > Lmin, different from edge at Lmin
            do     copy = Rangrid(Global->Rand, N);
            while( copy == edge );
            CALL( UserCopy(&Items[edge], &Items[copy]) )
            BoxCopy(Global, edge, copy);
            BoxCorr(Global, 2, edge, 0.5, copy, 0.5);
        }
        else
            BoxCorr(Global, 1, edge, 1.0);
//  explore
        CALL( UserExplore(Global, Items, L) )
        edge = BoxReset(Global);
//  terminate if  max contribution < accumulation / safety
        L = Global->lnLmax;
        if( tol > 0.0 &&  log(Number + N) + h + L < log(tol) + Evid )
            break;
    }

// Remainder steps
    a += log((N + 1.0) / (2.0 * N));
    for( i = 0; i < N; i++ )
    {
        Global->Nsteps++;
//  use each survivor
        L = Global->Lhood[i];
//  update system statistics
        b = PLUS(b, L + a);
        Evid  = PLUS(Evid,  h + L);
        Evid2 = PLUS(Evid2, b + L);
        b = PLUS(b, L + a);
        Global->wnew = wnew = exp(h + L - Evid);
        Global->wold = wold = 1.0 - wnew;
        Info = wold * Info + wnew * L;
        Logw = wold * Logw + wnew * (h + L);
        Number = exp(Evid - Logw);
        Global->lnEvidence  = 2.0 * Evid - 0.5 * Evid2;
        Global->devlnEvid   = sqrt(MAX(Evid2 - 2.0 * Evid, 0.0));
        Global->Information = Info - Evid;
        Global->Number  = (int)(0.5 + Number);
        Global->Nactive = (int)(0.5 + Number * Number / (PIE * N * N));
//  optional samples and user statistics
        CALL( BoxPosterior(Global, Samples, MaxSamples, &Items[i]) )
        CALL( UserProperties(Global, &Items[i], Samples) )
        BoxProperties(Global);
    }

Exit:
    FREE(Global->TopTree)
    FREE(Global->EdgeTree)
    FREE(Global->CorrTree)
    return CALLvalue;  
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function:  BoxCorr
// 
// Purpose:   Corr[item] := c, together with accompanying binary tree.
// 
// Method:    The BoxSampler system uses binary trees to ensure that its
//            internal workings remain O(log N) instead of the simpler O(N)
//            which could possibly dominate the CPU.
//            For example, N=5 correlation strengths Corr = {c0,c1,c2,c3,c4}
//            are stored as usual and can be scanned conventionally in O(N)
//            operations. Here the preceding Nplus=8 (power of 2 >= N) entries
//            hold partial sums that facilitate reaching specified places fast,
//            in O(log N) operations.  Blank values are 0.
//               -------------------------------------------------------
//     CorrTree |1                 c0+c1+c2+c3+c4+0+0+0                 |
//              |-------------------------------------------------------|
//              |2       c0+c1+c2+c3        |3        c4+0+0+0          |
//              |---------------------------|---------------------------|
//              |4   c0+c1    |5   c2+c3    |6   c4+0     |7    0+0     |
//              |-------------|-------------|-------------|-------------|
//        Corr->|8 c0  |9 c1  |10 c2 |11 c3 |12 c4 |13  0 |14  0 |15  0 |
//               -------------------------------------------------------
//            A similar tree for lnL, called TopTree, locates top values
//            by using MAX instead of addition; its blank values are ln(0).
//            Another tree for lnL, called EdgeTree, locates bottom values
//            by using MIN instead; its blank values are ln(infinity), and
//            has a low-order extension extTree which separates items whose
//            likelihoods happen to coincide.
//
// History:   John Skilling    3 April 2004
//-----------------------------------------------------------------------------
void BoxCorr(
GlobalStr* Global,  // I O  general information
int        ntry,    // I    # trial items with correlation strengths
 ... )              // I    item,c , item,c , item,c , ...
{
    double* CorrTree = Global->CorrTree;
    int     Nplus    = Global->Nplus;
    va_list ap;          // scan variable-argument list
    int     itry;        // counter for trial items
    int     item;        // alteration
    double  c;           // altered correlation strength
    int     j;           // tree address

    va_start(ap, ntry);
    for( itry = 0; itry < ntry; itry++ )
    {
        item = va_arg(ap, int);
        c    = va_arg(ap, double);
        j = Nplus + item;
        CorrTree[j] = c;
        for( ; j > 1; j >>= 1 )
            CorrTree[j>>1] = CorrTree[j] + CorrTree[j^1];
    }
    va_end(ap);
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function:  BoxLhood
// 
// Purpose:   Check if new items are all within box.  If OK, set
//                         Lhood[item] := lnLhood,
//            together with accompanying binary trees for min & max likelihood.
// 
// History:   John Skilling    3 April 2004
//-----------------------------------------------------------------------------
int BoxLhood(       //   O  1 if all inside box, otherwise 0 with no action
GlobalStr* Global,  // I O  general information
int        ntry,    // I    # trial items with lnLikelihood values
 ... )              // I    item,lnLhood , item,lnLhood , item,lnLhood , ...
{
    va_list ap;             // scan variable-argument list
    int     itry;           // counter for trial items
    int     item    = 0;    // replacement
    double  lnLhood = 0.0;  // replacement lnLikelihood
    double  ext     = 0.0;  // low-order extension of lnLikelihood value
    Rand_t  Rentry;         // recover input generator for re-use
    int     CALLvalue = 0;

    Rentry[0] = Global->Rand[0];
    Rentry[1] = Global->Rand[1];
    Rentry[2] = Global->Rand[2];
    Rentry[3] = Global->Rand[3];
    va_start(ap, ntry);
    for( itry = 0; itry < ntry; itry++ )
    {
        item    = va_arg(ap, int);
        lnLhood = va_arg(ap, double);
        ext     = Randouble(Global->Rand);
        if( GT( Global->lnLmin,Global->ext , lnLhood,ext ) ) // reject
            break;
    }
    if( itry == ntry )                                       // accept
    {
        va_end(ap);
        Global->Rand[0] = Rentry[0];
        Global->Rand[1] = Rentry[1];
        Global->Rand[2] = Rentry[2];
        Global->Rand[3] = Rentry[3];
        va_start(ap, ntry);
        for( itry = 0; itry < ntry; itry++ )
        {
            item    = va_arg(ap, int);
            lnLhood = va_arg(ap, double);
            ext     = Randouble(Global->Rand);
            BoxLtrees(item + Global->Nplus, lnLhood, ext,
                      Global->TopTree, Global->EdgeTree, Global->extTree);
        }
        CALLvalue = 1;
    }
    va_end(ap);
    return CALLvalue;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function:  BoxTarget
// 
// Purpose:   &(random item), weighted by correlation strengths.
// 
// History:   John Skilling    3 April 2004
//-----------------------------------------------------------------------------
int BoxTarget(      //   O  &(random item)
GlobalStr* Global)  // I    general information
{
    int       Nplus    = Global->Nplus;
    double*   CorrTree = Global->CorrTree;
    unsigned* Rand     = Global->Rand;
    double    r;    // random fraction of total strength
    int       j;

    r = CorrTree[1] * Randouble(Rand);
    for( j = 1; j < Nplus; )
    {
        j <<= 1;
        if( r > CorrTree[j] )
            r -= CorrTree[j++];
    }
    return j - Nplus;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function:  BoxCopy
// 
// Purpose:   Copy Lhood info onto edge item, after user has copied item itself
//            Set correlation strengths to 1/2 each.
// 
// History:   John Skilling    3 April 2004
//-----------------------------------------------------------------------------
static void BoxCopy(
GlobalStr* Global,    // I O  general information
int        edge,      // I    &(destination edge item to be over-written)
int        copy)      // I    &(source item to be duplicated)
{
    edge += Global->Nplus;
    copy += Global->Nplus;
    BoxLtrees(edge, Global->EdgeTree[copy], Global->extTree[copy],
              Global->TopTree, Global->EdgeTree, Global->extTree);
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function:  BoxLtrees
//
//            Set          Lhood[j] := lnLhood,
//            together with accompanying binary trees for min & max likelihood.
//
// History:   John Skilling    3 April 2004
//-----------------------------------------------------------------------------
static void BoxLtrees(
int      j,          // I    &(destination item) in trees
double   lnLhood,    // I    source ln Likelihood
double   ext,        // I    fraction of prior mass AT L deemed to be excluded
double*  TopTree,    // I O  binary tree for maximum likelihood
double*  EdgeTree,   // I O  binary tree for minimum likelihood
double*  extTree)    // I O  binary tree for likelihood extension
{
    EdgeTree[j] = TopTree[j] = lnLhood;
     extTree[j] = ext;
    for( ; j > 1; j >>= 1 )
    {
        if( GT( EdgeTree[j],extTree[j] , EdgeTree[j^1],extTree[j^1] ) )
        {
            EdgeTree[j>>1] = EdgeTree[j^1]; extTree[j>>1] = extTree[j^1];
        }
        else
        {
            EdgeTree[j>>1] = EdgeTree[j]; extTree[j>>1] = extTree[j];
        }
        TopTree[j>>1]  = MAX(TopTree[j],  TopTree[j^1]);
    }
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function:  BoxReset
// 
// Purpose:   Reset all correlation strengths to 0.
//            Find min and max lnLikelihood values.
//            Return item with min lnLikelihood.
// 
// History:   John Skilling    3 April 2004
//-----------------------------------------------------------------------------
static int BoxReset( //   O  &(outer item with min likelihood)
GlobalStr* Global)   // I    general information
{
    int     Nplus    = Global->Nplus;
    double* CorrTree = Global->CorrTree;
    double* TopTree  = Global->TopTree;
    double* EdgeTree = Global->EdgeTree;
    double* extTree  = Global->extTree;
    int     j;

    BoxReset1(1, Nplus, CorrTree);
    Global->lnLmax = TopTree[1];
    Global->lnLmin = EdgeTree[1];
    Global->ext   = extTree[1];

// Find edge item with min likelihood
    for( j = 1; j < Nplus; )
    {
        j <<= 1;
        if( GT( EdgeTree[j],extTree[j] , EdgeTree[j^1],extTree[j^1] ) )
           j++;
    }
    return j - Nplus;
}
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function:  BoxReset1
//            Recursive subsidiary of BoxReset
// History:   John Skilling    3 April 2004
//-----------------------------------------------------------------------------
static void BoxReset1(int i, int Nplus, double* Corr)
{
    int j = i;
    if( j < Nplus )
    {
        j <<= 1;
        if( Corr[j]   > 0.0 )  BoxReset1(j,   Nplus, Corr);
        if( Corr[j^1] > 0.0 )  BoxReset1(j^1, Nplus, Corr);
    }
    Corr[i] = 0.0;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function:  BoxPosterior
// 
// Purpose:   Update posterior ensemble with new candidate item, ensuring
//            that all multiplicities stay <= 1 to avoid duplication.
// 
// History:   John Skilling    3 April 2004
//-----------------------------------------------------------------------------
static 
int BoxPosterior(       //   O  >=0 OK, or -ve error code
GlobalStr* Global,      // I    general information
ItemStr*   pSamples,    // I O  posterior samples
int        MaxSamples,  // I    max # samples
ItemStr*   Item)        // I    candidate item
{
    unsigned* Rand = Global->Rpost;
    double    curr;     // current membership, as floating point
    double    next;     // incremental membership, as floating point
    double    r;        // random, for digitising memberships
    int       keep;     // # old items kept, <= Nsamples
    int       insert;   // # copies of new item, 0 or 1
    int       kill;     // for removing surplus items
    int       n;        // local copy of Nsamples
    int       CALLvalue = 0;

    if( MaxSamples > 0 && pSamples )
    {
        n = Global->Nsamples;
// # copies
        if( n <= 0 )
        {
            keep   = 0;
            insert = 1;
        }
        else
        {
// Fit into max size
            curr = MaxSamples * Global->wold;
            next = MaxSamples * Global->wnew;
// Next weight must be <= 1
            if( next > 1.0 )
            {
                curr /= next;
                next = 1.0;
            }
// Old weights must be <= 1
            if( curr > n )
            {
                next *= n / curr;
                curr = n;
            }
// # copies
            r = Randouble(Rand);
            insert = (int)(r + next);
            keep   = (int)(r + next + curr) - insert;
        }
// Guard against silly rounding
        while( keep + insert > MaxSamples )
            keep--;
// Kill random different input items, leaving only keep
        while( n > keep )
        {
            kill = Rangrid(Rand, n--);
            CALL( UserCopy(&pSamples[kill], &pSamples[n]) )
        }
// Optional new item
        if( insert )
        {
            CALL( UserCopy(&pSamples[keep], Item) )
            ++ n;
        }
        Global->Nsamples = n;
    }
Exit:
    return CALLvalue;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function:  BoxProperties
// 
// Purpose:   Update mean and std.dev. of properties.
// 
// History:   John Skilling    3 April 2004
//-----------------------------------------------------------------------------
static void BoxProperties(
GlobalStr* Global)                    // I O  general information
{
    int     n = Global->NumProps;     // I    # properties
    double  wnew = Global->wnew;      // I    proportional weight of new item
    double  wold = Global->wold;      // I    proportional weight of old items
    double* Prop = Global->Properties;// I    user's properties
    double* Mean = Global->Mean;      //   O  updated mean
    double* Dev  = Global->Dev;       //   O  updated standard deviation
    double  x;
    int     i;

    for( i = 0; i < n; i++ )
    {
        x = Prop[i] - Mean[i];                // offset
        Mean[i] += wnew * x;                  // update mean
        Dev[i] = sqrt(wold * (wnew * x * x + Dev[i] * Dev[i]));
    }                                         // non-negative variance update
}
