//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//            Box Sampling
// 
// Filename:  box.h
// 
// Purpose:   Header for box.c.  Documents user-accessible procedures.
// 
// History:   JS    3 April 2004
//-----------------------------------------------------------------------------
#ifndef BOXSAMPLER
#define BOXSAMPLER
#include "userstr.h"       // ensure ItemStr is defined
#include "random.h"        // ensure random procedures are accessible

/**************/
/* Structures */
/**************/

typedef struct             // <<<< SYSTEM INFORMATION >>>>
{
    double   lnEvidence;   //   O  ln(Evidence)
    double   devlnEvid;    //   O  standard deviation of computed ln(Evidence)
    double   Information;  //   O  <ln(posterior)> = ln(compression factor)
    int      Iseed;        // I O  Random seed, +ve = fixed, -ve = time seed
    int      NumItems;     // I    # items in evolving collection
    int      NumProps;     // I    # properties to be accumulated
    double*  Properties;   // I    properties to be quantified      [NumProps]
    double*  Mean;         //   O  mean properties                  [NumProps]
    double*  Dev;          //   O  standard deviation properties    [NumProps]
    int      Nsteps;       //   O  total # items computed and used
    int      Nsamples;     //   O  actual # posterior samples (if generated)
    int      Number;       //   O  effective # independent samples during run
    int      Nactive;      //   O  effective # active constraints
    double   wnew;         //  (O) fraction of new mean assigned to new item
    double   wold;         //  (O) old contribution = 1 - wnew
    int      Nplus;        //  (O) power-of-2 >= NumItems
    double   lnLmax;       //  (O) top  (maximum) lnLikelihood
    double   lnLmin;       //  (O) edge (minimum) lnLikelihood
    double   ext;          //  (O) low-order extension of edge lnLikelihood
    double*  Lhood;        //  (O) lnLikelihood                     [NumItems]
    double*  Corr;         //  (O) correlation strengths            [NumItems]
    double*  CorrTree;     //  (O) binary tree for correlation strengths
    double*  TopTree;      //  (O) binary tree for maximum lnLikelihood
    double*  EdgeTree;     //  (O) binary tree for minimum lnLikelihood
    double*  extTree;      //  (O) binary tree for low-order extension to min
    Rand_t   Rand;         //  (O) random generator for general use
    Rand_t   Rpost;        //  (O) separate random generator for posterior
} GlobalStr;

/**************/
/* Procedures */
/**************/
#ifdef __cplusplus
extern "C" {
#endif

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function:  BoxSampler
// 
// Purpose:   Bayesian computation
//
//                   Pr(x)  Pr(D|x)   =    Pr(D)  Pr(x|D)
//                  Prior Likelihood ==> Evidence Posterior
//
//            Use "BoxSampler" to obtain
//
//               lnEvidence  := ln INTEGRAL Likelihood d(Prior),
//               devlnEvid   := std.dev. uncertainty of lnEvidence
//
//               Information := ln( prior-to-posterior compression factor )
//            
//               Number      := effective # independent samples during run
//
//               Nactive     := apparent # active constraints
//
//               Posterior samples (optional)
//
//               Posterior mean properties (optional)
//-----------------------------------------------------------------------------
extern int  BoxSampler(    //   O  UserProperties return code, or -ve error
    GlobalStr* Global,     // I O  general information
    ItemStr*   Items,      //   O  item collection           [Global->NumItems]
    ItemStr*   Samples,    //  (O) posterior samples            [<= MaxSamples]
    int        MaxSamples, // I    max # posterior samples      (0 allowed)
    int        MaxSteps,   // I    max # steps allowed        (-ve is inactive)
    double     tol);       // I    tolerance on Evidence        (0 is inactive)

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function:  UserPrior
// 
// Purpose:   Set one item sampled from the prior alone.
//-----------------------------------------------------------------------------
extern int UserPrior(      //   O  >=0 OK, or -ve error code
    GlobalStr* Global,     // I    general information
    ItemStr*   Item,       //   O  item being set
    double*    lnLhood);   //   O  lnLikelihood

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function:  UserExplore
// 
// Purpose:   Re-randomise collection of items.
//
//            I supply you with a collection of  N = Global->NumItems  items.
//            Though you may not need to know, their lnLikelihoods are in
//                    Global->Lhood[0...N-1].
//
//            Assuming N > 1, I should now be supplying you with each item
//            sampled uniformly (with respect to the prior) within the box
//                   "lnLikelihood >= lnLmin".
//            Specifically, items less probable than Lmin are excluded,
//            items more probable are included, and I control whether or not 
//            items exactly at Lmin are to be included.
//            Items will all be independent EXCEPT that 2 of them are the same.
//            These two ("s" and "t") are the ones with correlation strengths
//                    Global->Corr[s] = Global->Corr[t] = 1/2
//            all other items having zero correlation.
//            (If N is only 1, I give you a single item on the box edge).
//
//            Please do the best you can to smooth away the correlation and
//            make ALL the items INDEPENDENT while keeping the items uniform
//            within the "lnLikelihood >= lnLmin" box.
//            You can use any method you choose, e.g.
//        (1) Re-sample s or t ab initio leaving all others intact;
//        (2) MCMC exploration of item s and/or t, possibly using other items
//            as stationary guides;
//        (3) Genetic mixing of any items you wish.
//            Your method must remain faithful to the prior within the box,
//            but can allow trial transitions to outside.
//
//            When you have n trial item(s) i,j,... with their lnLikelihood
//            values Li,Lj,..., accept them if and only if
//                    BoxLhood(Global, n, i,Li, j,Lj, ... ) != 0
//            This checks that the likelihood values are all correctly
//            within the box, including paying attention to the exact edge.
//            If they are, BoxLhood then resets the binary trees on which
//            my updating relies, and also has the obvious effect
//                   "Global->Lhood[i] = Li", ...
//            for each of your n trial items.
//            Do not just reset Global->Lhood yourself.
//
//            You may want to target an item randomly according to its
//            correlation strength, which you can do in O(log N) with
//                    target = BoxTarget(Global);
//            Initially this can only give you s or t.
//            If you want to use genetic mixing, you must re-distribute
//            the correlation strengths of the mixed items accordingly
//            (preserving the sum of the strengths).  Otherwise you risk
//            leaving an indelible trail of correlation which is against
//            our purpose.  To reset correlation strengths when accepting
//            trial items, use
//                    BoxCorr(Global, n, i,ci, j,cj, ... );
//            This resets my binary tree as well as having the obvious effects
//                   "Global->Corr[i] = ci", ...
//            Do not just reset Global->Corr yourself.
//-----------------------------------------------------------------------------
extern int UserExplore(    //   O  >=0 OK, or -ve error code
    GlobalStr* Global,     // I    general information
    ItemStr*   Items,      // I O  collection of items      [Global->NumItems]
    double     lnLmin);    // I    minimum bounding lnLikelihood

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function:  UserCopy
// 
// Purpose:   Copy item,  " Dest := Src ".
//-----------------------------------------------------------------------------
extern int UserCopy(       //   O  >=0 OK, or -ve error code
    ItemStr*   Dest,       //   O  destination, to be over-written
    ItemStr*   Src);       // I    input item, to be copied

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function:  UserProperties
// 
// Purpose:   Item is the item on the edge, to be used now before overwritten.
//            Its lnLikelihood is in Global->lnLmin.
//
//            Calculate any of its Properties that are to be quantified.
//
//            You may also attend to diagnostics here.
//-----------------------------------------------------------------------------
extern int UserProperties( //   O  >=0 OK, or -ve error code
    GlobalStr* Global,     // I    general information, incl. Properties list
    ItemStr*   Item,       // I    item about to be used
    ItemStr*   Samples);   // I    posterior samples         [Global->Nsamples]

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function:  BoxCorr
// 
// Purpose:   Corr[item] := c, together with accompanying binary tree.
//-----------------------------------------------------------------------------
extern void BoxCorr(
    GlobalStr* Global,     // I O  general information
    int        ntry,       // I    # trial items with correlation strengths
    ... );                 // I    item,c , item,c , ...

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function:  BoxLhood
// 
// Purpose:   Check if new items are all within box.  If OK, set
//                         Lhood[item] := lnLhood,
//            together with accompanying binary trees for min & max likelihood.
//-----------------------------------------------------------------------------
extern int  BoxLhood(      //   O  1 if inside box, otherwise 0 with no action
    GlobalStr* Global,     // I O  general information
    int        ntry,       // I    # trial items with lnLikelihood values
    ... );                 // I    item,lnLhood , item,lnLhood , ...

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function:  BoxTarget
// 
// Purpose:   &(random item), weighted by correlation strengths.
//-----------------------------------------------------------------------------
extern int  BoxTarget(     //   O  &(random item from correlation strengths)
    GlobalStr* Global);    // I    general information

#ifdef __cplusplus
};
#endif

/*************/
/* Constants */
/*************/
#undef  E_MALLOC     // Memory allocation error
#undef  E_RAN_ARITH  // Random generator requires 32 bits integer precision

#define E_MALLOC     -130
#define E_RAN_ARITH  -299

#endif

