//=============================================================================
//
//                  BOX SAMPLING EXAMPLE PROGRAM  sphere.c
//
// Link
//    sphere.c = {main,UserPrior,UserExplore,UserCopy,UserProperties}   *USER*
//    userstr.h = {definition of ItemStr, def of your error codes}      *USER*
// with
//    box.c, box.h                                                    *SYSTEM*
//    random.c, random.h                                              *SYSTEM*
//=============================================================================
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "userstr.h"              // Your header, needed before box.h
#include "box.h"                  // My header, needs definitions in userstr.h
#include "random.h"               // Random library

int main()
{
// System parameters
    int       NumItems   = 10;    // # items in internal collection
    int       MaxSamples = 10;    // max # posterior Samples (can be 0)
    int       MaxSteps   = -1;    // +ve is limit (-ve is inactive)
    double    tol        = 1e-6;  // numerical tolerance (0 is inactive)
    int       Iseed      = 4321;  // random seed (-ve is time seed)
// User application parameters and/or Data
    int       Ndim       = 40;    // application is 40-dimensional,...
    double    Radius     = 50.0;  // ... within a sphere of radius 50.

    GlobalStr Global[1];  // system information
    ParamStr  Params[1];  // common parameters
    ItemStr*  Items;      // individual items in working collection [NumItems]
    ItemStr*  Samples;    // posterior Samples, if needed         [MaxSamples]
    int       code;       // BoxSampler return code (-ve is error)
    double    Z;          // true lnEvidence, for comparison
    int       j;          // counter

// System
    Global->NumItems   = NumItems;
    Global->NumProps   = 2;
    Global->Properties = malloc(Global->NumProps * sizeof(double));
    Global->Mean       = malloc(Global->NumProps * sizeof(double));
    Global->Dev        = malloc(Global->NumProps * sizeof(double));
    Global->Iseed      = Iseed;
// User application
    Params->Ndim   = Ndim;
    Params->Radius = Radius;
    Params->try    = malloc(Ndim * sizeof(double));
// Create Items and (optionally) Samples
    Items = malloc(NumItems * sizeof(ItemStr));
    for( j = 0; j < NumItems; j++ )
    {
        Items[j].Params = Params;
        Items[j].x = malloc(Ndim * sizeof(double));
    }
    Samples = malloc(MaxSamples * sizeof(ItemStr));
    for( j = 0; j < MaxSamples; j++ )
    {
        Samples[j].Params = Params;
        Samples[j].x = malloc(Ndim * sizeof(double));
    }

// Run >>>>>>>>>>>
    code = BoxSampler(Global, Items, Samples, MaxSamples, MaxSteps, tol);
// <<<<<<<<<<<<<<<

// Results
    printf("\nReturn code = %d\n", code);
    if( code < 0 )
        printf("\a ERROR STATE\n");
    else
    {
        Z = (Ndim & 1) ? 0.5 * log(3.14159265 / (2.0 * Radius * Radius)) : 0.0;
        for( j = Ndim; j > 1; j -= 2 )
            Z += log(j / (Radius * Radius));
        printf("Seed        = %d\n", Global->Iseed);
        printf("Nsteps      = %d\n", Global->Nsteps);
        printf("lnEvidence  =%10.4f +-%7.4f  ",
                Global->lnEvidence, Global->devlnEvid);
        printf("  should be  %10.4f\n", Z);
        printf("Information =%10.4f            ", Global->Information);
        printf("  should be  %10.4f\n", -0.5 * Ndim - Z);
        printf("Effective # active constraints = %d", Global->Nactive);
        printf("  should be  %d\n", Ndim);
        printf("Effective # independent items  = %d\n", Global->Number);
// Sample coordinates should be from unit Gaussian in this application
        for( j = 0; j < Global->Nsamples; j++ )
            printf("    Sample%2d x[0] = %7.4f\n",  j, Samples[j].x[0]);
// Statistics are accumulated during the run better than from the samples
        printf("Rsquared =%9.4f +-%8.4f", Global->Mean[0], Global->Dev[0]);
        printf("  should be %4d.0000 +-%8.4f\n", Ndim, sqrt(2.0 * Ndim));
        printf("  x[0]   =%9.4f +-%8.4f", Global->Mean[1], Global->Dev[1]);
        printf("  should be    0 +- 1\n");
    }
    return 0;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function:  UserPrior
// 
// Purpose:   Set one item sampled from the prior alone.
//
// Example:   This code uses flat prior in Ndim-dimensional sphere of
//            radius Radius and volume (pi Radius^2)^(Ndim/2) / (Ndim/2)!
//            Likelihood is Gaussian: at intermediate radius r,
//                lnLikelihood = - r^2 / 2
//
//            This has sufficient symmetry to enable exact ab initio sampling.
//            Correct results for large Radius (> sqrt(Ndim)+3 suffices) are
//                lnEvidence  = ln( 2^(Ndim/2) (Ndim/2)! / Radius^Ndim )
//                Information = - Ndim/2 - lnEvidence
//            e.g. Ndim = 40,  Radius = 50,
//                lnEvidence  = -100.2824
//                Information =   80.2824, a factor exp(80.2824) compression
// 
// History:   JS        3 April 2004
//-----------------------------------------------------------------------------
int UserPrior(               //   O  >=0 OK, or -ve error code
GlobalStr* Global,           // I    general information
ItemStr*   Item,             //   O  item being set
double*    lnLhood)          //   O  lnLikelihood
{
    double*   x      = Item->x;         // shortcut to coordinates
    ParamStr* Params = Item->Params;    // common parameters
    int       Ndim   = Params->Ndim;    // dimension
    double    a      = Params->Radius;  // prior (maximum) radius
    unsigned* Rand   = Global->Rand;    // random generator
    double    rr;                       // local radius^2
    double    s;                        // scale factor
    int       i;

    if( a < sqrt(Ndim) + 3.0 )
        return E_ANALYSIS;    // example of abort with user's -ve return code
// New radius enclosing fractional volume from Uniform[0,1]
    a *= pow(Randouble(Rand), 1.0 / Ndim);
// Sample from unit Gaussian ball
    for( i = 0; i < Ndim; i++ )
        x[i] = Rangauss(Rand);
    rr = 0.0;    for( i = 0; i < Ndim; i++ )    rr += x[i] * x[i];
// Scale onto new radius
    s = a / sqrt(rr);    for( i = 0; i < Ndim; i++ )    x[i] *= s;
// New lnLikelihood
    *lnLhood = -0.5 * a * a;
    return  0;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function:  UserExplore
// 
// Purpose:   Re-randomise item(s)
//
// History:   JS        3 April 2004
//-----------------------------------------------------------------------------
int UserExplore(             //   O  >=0 OK, or -ve error code
GlobalStr* Global,           // I    general information
ItemStr*   Items,            // I O  collection of items     [Global->NumItems]
double     lnLmin)           // I    minimum bounding lnLikelihood
{
    ParamStr* Params = Items->Params;   // common information
    int       Ndim   = Params->Ndim;    // dimension
    double*   try    = Params->try;     // trial position
    unsigned* Rand   = Global->Rand;    // random generator
    int       target;                   // single item moved in this example
    int       MCMC;                     // Markov chain Monte Carlo counter
    double    s;                        // Markov chain Monte Carlo step-size
    double    rr;                       // trial radius^2
    double    radius;                   // box limit
    double    L;                        // ln likelihood
    double*   x;                        // shortcut to coordinates
    int       i;

// Radius of box
    radius = sqrt(-2.0 * lnLmin);
#if 0                              // EITHER EXACT AB INITIO SAMPLING >>>>>>>>

// Select single item to be moved, randomly by correlation strength
    target = BoxTarget(Global);                     // RECOMMENDED SYSTEM CALL
    x = Items[target].x;
// New item radius
    radius *= pow(Randouble(Rand), 1.0 / Ndim);
// Sample from unit Gaussian ball
    for( i = 0; i < Ndim; i++ )
        x[i] = Rangauss(Rand);
    rr = 0.0;    for( i = 0; i < Ndim; i++ )    rr += x[i] * x[i];
// Scale onto new radius
    s = radius / sqrt(rr);    for( i = 0; i < Ndim; i++ )    x[i] *= s;
// Use new item, guaranteed inside box
    L = -0.5 * radius * radius;
    BoxLhood(Global, 1, target, L);                 // REQUIRED SYSTEM CALL

#else                             // OR MCMC APPROXIMATION >>>>>>>>>>>>>>>>>>>

// MCMC step length, optimised for speed
    s = radius * 1.8507 / Ndim;
// Guess number of steps
    for( MCMC = 0; MCMC < 40; MCMC++ )
    {
// Select single item to be moved, randomly by correlation strength
        target = BoxTarget(Global);                 // RECOMMENDED SYSTEM CALL
        x = Items[target].x;
// Trial item
        for( i = 0; i < Ndim; i++ )
            try[i] = x[i] + s * Rangauss(Rand);
        rr = 0.0;    for( i = 0; i < Ndim; i++ )    rr += try[i] * try[i];
// Accept/reject: if OK, use trial item
        L = -0.5 * rr;
        if( BoxLhood(Global, 1, target, L) )        // REQUIRED SYSTEM CALL
        {
            for( i = 0; i < Ndim; i++ )
                x[i] = try[i];
        }
    }

#endif                            // <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
    return  0;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function:  UserCopy
// 
// Purpose:   Copy item,  " Dest := Src ".
// 
// History:   JS        3 April 2004
//-----------------------------------------------------------------------------
int UserCopy(                //   O  >=0 OK, or -ve error code
ItemStr*   Dest,             //   O  destination, to be over-written
ItemStr*   Src)              // I    input item, to be copied
{
    int       Ndim   = Src->Params->Ndim;
    double*   src    = Src->x;
    double*   dest   = Dest->x;
    int       i;

    for( i = 0; i < Ndim; i++ )
        dest[i] = src[i];
    return 0;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function:  UserProperties
// 
// Purpose:   Properties and diagnostics.
//
// History:   JS        3 April 2004
//-----------------------------------------------------------------------------
int UserProperties(          //   O  >=0 OK, or -ve error code
GlobalStr* Global,           // I    general information, incl. Properties list
ItemStr*   Item,             // I    item, with lnLikelihood = Global->lnLmin
ItemStr*   Samples)          // I    posterior samples       [Global->Nsamples]
{
    int       i;

// Set desired properties, e.g.....
    Global->Properties[0] = -2.0 * Global->lnLmin;  // radius-squared
    Global->Properties[1] = Item->x[0];             // one of the coordinates

// Arbitrary diagnostics
    printf("\nStep#%6d    lnEvid = %8.2f     lnLhood = %8.2f\n",
              Global->Nsteps, Global->lnEvidence, Global->lnLmin);
    for( i = 0; i < Global->Nsamples; i++ )
        printf("    Sample%2d x[0] = %7.4f\n",  i, Samples[i].x[0]);

    return  0;
}
